# $Id: card.rb,v 1.2 2005/11/24 12:20:45 noir Exp $
# Copyright (C) 2005 Hiroyuki KUROSAKI <noir (at) st . rim . or . jp>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# This file is required by simplezdb.so.

module SimpleZDB
  class Card
    include Enumerable

# call-seq:
#   Card.new(database, card_id) -> card
#
# Create a new Card object. You should specify valid DataBase object
# and card_id (Integer).
# If card_id is 0, it is treated as a new card.

    def initialize(database, card_id)
      @database = database
      @card_id  = card_id
      @values_cache = Hash.new
    end

    attr_reader :database
    attr_reader :card_id

    alias :to_ary :to_a

# call-seq:
#   card.keys -> array
#
# Return keys of the card.

    def keys
      @database.item_info.names
    end
      
# call-seq:
#   card[key] -> object
#
# Return the item associated with key.
# The type of the item is determined by the corresponding element
# of database.item_info.

    def [](key)
      return @values_cache[key] if @values_cache[key]
      val = @database.read_item(key, @card_id)
      @values_cache[key] = val
      val
    end

# call-seq:
#   card[key] = val
#
# Assign the value associated with the key of the card.
# Values are stored temporary on memory.
# To save data, call card.save.
# The type of the value should correspond to the element
# of database.item_info.

    def []=(key, val)
      @values_cache[key] = val
    end

# call-seq:
#   card.save -> true or false
#
# Save the temporary data to database file.

    def save
      result = @database.start_edit_card(@card_id)
      unless result
        raise RuntimeError, "Start editing card #{@card_id} failed."
        return false
      end

      self.keys.each do |key|
        val = @values_cache[key]
        next unless val
        result = @database.write_item(key, val)
        unless result
          @database.cancel_edit_card
          raise RuntimeError, "Save item #{key} failed."
          return false
        end
      end

      new_card_id = @database.finish_edit_card(@card_id)
      unless @card_id
        raise RuntimeError, "Finish editing card #{@card_id} failed."
        return false
      end

      @card_id = new_card_id

      cats = self.categories

      if cats && cats.size > 0 then
        result = @database.update_categories(cats, @card_id)
        unless result
          raise RuntimeError, "Updating categories of #{@card_id} failed."
          return false
        end
      end

      true
    end

# call-seq:
#   card.delete -> true or false
#
# Delete this card.

    def delete
      @database.delete_card(@card_id)
    end

# call-seq:
#   card.each {|val| ... } -> card
#
# Iterate the values of the card. The order is the same as of the
# item_info.names of the database.

    def each
      @database.item_info.names.each do |name|
        yield(self[name])
      end
      self
    end

# call-seq:
#   card.to_hash -> hash
#
# Return key-value pairs of the card as hash.

    def to_hash
      tmp_hash = Hash.new
      @database.item_info.names.each do |name|
        tmp_hash[name] = self[name]
      end
      tmp_hash
    end

# call-seq:
#   card.categories -> array
#
# Return categories of the card.

    def categories
      self["CATEGORY"]
    end

# call-seq:
#   card.categories = ["string", "string", ...]
# 
# Assign the categories of the card.

    def categories=(new_cats)
      self["CATEGORY"] = new_cats
    end

# call-seq:
#   Card.define_accessors(keys)
#
# define accessors named by hash key.

    def self.define_accessors(keys)
      unless method_defined?(keys.first.downcase.to_sym)
        keys.each do |key|
          define_method(key.downcase.to_sym) {
            self[key]
          }
          define_method("#{key.downcase}=".to_sym) {|val|
            self[key] = val
          }
        end
      end
    end

  end

end
