/*
  simplezdb.cpp - a ruby extension module providing the simple access to ZDB
  $Id: simplezdb.cpp,v 1.13 2005/11/25 22:20:25 noir Exp $
  Copyright (C) 2005 Hiroyuki KUROSAKI <noir (at) st . rim . or . jp>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ruby.h"

#include <sl/slzdb.h>
#include <qdatetime.h>

static VALUE mSimpleZDB;
static VALUE cItemInfo;
static VALUE cSortKey;
static VALUE cDataBase;
static VALUE cCategorySet;

VALUE cCard;
VALUE cDate, cDateTime;

/* Generic functions */

static VALUE each(VALUE obj)
{
    return rb_funcall(obj, rb_intern("each"), 0);
}

/*
 * Document-class: SimpleZDB::ItemInfo
 *
 * Almost useless as a wrapper of original SlZDataBaseItemInfo,
 * since we cannot refer to its members.
 * It also aggregates names, value_types, titles, is_searchables.
 * (titles and is_searchables are not available for now.)
 * The index number of each array means "field" for read/write methods
 * of SlZDataBase.
 * Index 0 is always categories.
 */

static const char* ITEMINFO_TYPES[] = {
    "TEXT",
    "UCHAR",
    "USHORT",
    "ULONG",
    "COLOR",
    "TIME",
    "BINARY",
    "CATEGORY",
};

static void sli_free(void* p)
{
    delete static_cast<SlZDataBaseItemInfo*>(p);
}

static VALUE sli_alloc(VALUE klass)
{
    SlZDataBaseItemInfo* p_sli;
    VALUE obj;

    p_sli = new SlZDataBaseItemInfo();
    obj = Data_Wrap_Struct(klass, 0, sli_free, p_sli);
    return obj;
}

SlZDataBaseItemInfo* ite_get_sli(VALUE self)
{
    SlZDataBaseItemInfo* p_sli;
    Data_Get_Struct(self, SlZDataBaseItemInfo, p_sli);
    return p_sli;
}

/*
 * Document-method: ItemInfo.new
 * call-seq:
 *   ItemInfo.new
 *
 * Create ItemInfo instance.
 */
static VALUE ite_initialize(VALUE self)
{
    VALUE names          = rb_ary_new();
    VALUE value_types    = rb_ary_new();
    VALUE titles         = rb_ary_new();
    VALUE is_searchables = rb_ary_new();

    rb_ary_push(names,          rb_str_new2("CATEGORY"));
    rb_ary_push(value_types,    rb_const_get(cItemInfo, rb_intern("CATEGORY")));
    rb_ary_push(titles,         Qnil);
    rb_ary_push(is_searchables, Qfalse);

    rb_iv_set(self, "@names",          names);
    rb_iv_set(self, "@value_types",    value_types);
    rb_iv_set(self, "@titles",         titles);
    rb_iv_set(self, "@is_searchables", is_searchables);
    return self;
}

/*
 * Document-method: SimpleZDB::ItemInfo#each
 * call-seq:
 *   item_info.each {|field, name, value_type| ... } -> item_info
 *
 * Iterate the ItemInfo.
 */
static VALUE ite_each(VALUE self)
{
    int size = NUM2INT(rb_funcall(rb_iv_get(self, "@names"),
                                  rb_intern("size"),
                                  0));
    VALUE names       = rb_iv_get(self, "@names");
    VALUE value_types = rb_iv_get(self, "@value_types");
    for (int i = 0; i < size; ++i) {
        VALUE field = INT2FIX(i);
        rb_yield_values(3,
                        field,
                        rb_funcall(names, rb_intern("[]"), 1, field),
                        rb_funcall(value_types, rb_intern("[]"), 1, field));
    }
    return self;
}

/*
 * Document-method: add_item_info
 * call-seq:
 *   item_info.add_item_info(field, item_name, item_type, title, is_searchable)
 *
 * Wrapper of SlZDataBaseItemInfo::addItemInfo(int, const char*, ItemType, const QString, bool).
 * Possible values of item_type are:
 *
 *    ItemInfo::TEXT
 *    ItemInfo::UCHAR
 *    ItemInfo::USHORT
 *    ItemInfo::ULONG
 *    ItemInfo::COLOR
 *    ItemInfo::TIME
 *    ItemInfo::BINARY
 *    ItemInfo::CATEGORY
 *
 * Besides, add each value to the instance variables names, value_types,
 * titles, is_searchables.
 */

static VALUE ite_add_item_info(VALUE self, VALUE args)
{
    int argc = NUM2INT(rb_funcall(args, rb_intern("size"), 0));
    if (argc < 4) {
        rb_raise(rb_eRuntimeError, "Argument too few (%d of 4 or 5)");
        return Qnil;
    }
    VALUE v_field           = rb_ary_shift(args);
    VALUE v_orig_item_name  = rb_ary_shift(args);
    VALUE v_item_type       = rb_ary_shift(args);
    VALUE v_orig_title      = rb_ary_shift(args);
    VALUE v_is_searchable   = rb_ary_shift(args);
    if (NIL_P(v_orig_title)) {
        v_orig_title = rb_str_new2("(no title)");
    }
    if (NIL_P(v_is_searchable)) {
        v_is_searchable = Qtrue;
    }
    VALUE v_item_name = StringValue(v_orig_item_name);
    VALUE v_title     = StringValue(v_orig_title);
    QString q_title(StringValuePtr(v_title));

    SlZDataBaseItemInfo* p_sli = ite_get_sli(self);
    p_sli->addItemInfo(NUM2INT(v_field),
                       StringValuePtr(v_item_name),
                       (SlZDataBaseItemInfo::ItemType)NUM2INT(v_item_type),
                       q_title,
                       RTEST(v_is_searchable));

    rb_funcall(rb_iv_get(self, "@names"),
               rb_intern("[]="),
               2,
               v_field,
               v_item_name);
    rb_funcall(rb_iv_get(self, "@value_types"),
               rb_intern("[]="),
               2,
               v_field,
               v_item_type);
    rb_funcall(rb_iv_get(self, "@titles"),
               rb_intern("[]="),
               2,
               v_field,
               v_title);
    rb_funcall(rb_iv_get(self, "@is_searchables"),
               rb_intern("[]="),
               2,
               v_field,
               v_is_searchable);
    
    return Qnil;
}

/*
 * Document-method: SimpleZDB::ItemInfo#field(key)
 * call-seq:
 *   item_info.field(name)
 *
 * Return the field (i.e, index) number of the corresponding name.
 */
static VALUE ite_field(VALUE self, VALUE name)
{
    return rb_funcall(rb_iv_get(self, "@names"),
                      rb_intern("index"),
                      1,
                      name);
}
/*
 * Document-method: SimpleZDB::ItemInfo#value_type(key)
 * call-seq:
 *   item_info.value_type(name)
 *
 * Return the value type of the corresponding name.
 */
static VALUE ite_value_type(VALUE self, VALUE name)
{
    return rb_funcall(rb_iv_get(self, "@value_types"),
                      rb_intern("[]"),
                      1,
                      ite_field(self, name));
}

/*
 * Document-method: SimpleZDB::ItemInfo#key(field)
 * call-seq:
 *   item_info.name(field)
 *
 * Return the name corresponding the field (i.e., index) number.
 */
static VALUE ite_name(VALUE self, VALUE field)
{
    return rb_funcall(rb_iv_get(self, "@names"),
                      rb_intern("[]"),
                      1,
                      field);
}

/* Document-method: SimpleZDB::ItemInfo#size
 * call-seq:
 *   item_info.size
 *
 * Return the size of the ItemInfo.
 */
static VALUE ite_size(VALUE self)
{
    return rb_funcall(rb_iv_get(self, "@names"),
                      rb_intern("size"),
                      0);
}

/* Document-method: SimpleZDB::ItemInfo#include?
 * call-seq:
 *   item_info.include?(name) -> true or false
 *
 * If item_info includes the name, return true.
 */
static VALUE ite_include_p(VALUE self, VALUE name)
{
    return rb_funcall(rb_iv_get(self, "@names"),
                      rb_intern("include?"),
                      1,
                      name);
}


/*
 * Document-class: SimpleZDB::SortKey
 *
 * Wrapper of SlZDataManagerSortKey class.
 */
static void zsk_free(void* p)
{
    delete static_cast<SlZDataManagerSortKey*>(p);
}

static VALUE zsk_alloc(VALUE klass)
{
    SlZDataManagerSortKey* p_zsk;
    VALUE obj;

    p_zsk = new SlZDataManagerSortKey();
    obj = Data_Wrap_Struct(klass, 0, zsk_free, p_zsk);
    return obj;
}

SlZDataManagerSortKey* sk_get_zsk(VALUE self)
{
    SlZDataManagerSortKey* p_zsk;
    Data_Get_Struct(self, SlZDataManagerSortKey, p_zsk);
    return p_zsk;
}

/*
 * Document-method: SimpleZDB::SortKey.new
 *
 * Instanciate SortKey object.
 */
static VALUE sk_initialize(VALUE self)
{
    VALUE item_names    = rb_ary_new();
    VALUE is_strings    = rb_ary_new();
    VALUE is_ascendents = rb_ary_new();
    rb_iv_set(self, "@item_names",    item_names);
    rb_iv_set(self, "@is_strings",    is_strings);
    rb_iv_set(self, "@is_ascendents", is_ascendents);
    return self;
}

/*
 * Document-method: add_sort_key
 * call-seq:
 *   sort_key.add_sort_key(item_name, is_string, is_ascendent) -> nil
 *
 * Wrapper of SlZDataManagerSortKey::addSortKey(const char*, bool, bool).
 */
static VALUE sk_add_sort_key(VALUE self, VALUE args)
{
    VALUE orig_item_name = rb_ary_shift(args);
    VALUE is_string      = rb_ary_shift(args);
    VALUE is_ascendent   = rb_ary_shift(args);
    VALUE item_name;
    if (NIL_P(item_name))
        return Qnil;
    else
        item_name = StringValue(orig_item_name);
    if (NIL_P(is_string))
        is_string = Qfalse;
    else
        is_string = RTEST(is_string) ? Qtrue : Qfalse;
    if (NIL_P(is_ascendent))
        is_ascendent = Qtrue;
    else
        is_ascendent = RTEST(is_ascendent) ? Qtrue : Qfalse;

    SlZDataManagerSortKey* p_zsk = sk_get_zsk(self);

    p_zsk->addSortKey(StringValuePtr(item_name),
                      RTEST(is_string),
                      RTEST(is_ascendent));

    rb_ary_push(rb_iv_get(self, "@item_names"),    item_name);
    rb_ary_push(rb_iv_get(self, "@is_strings"),    is_string);
    rb_ary_push(rb_iv_get(self, "@is_ascendents"), is_ascendent);
    return Qnil;
}

/* Document-method: create_sort_exp
 * call-seq:
 *   sort_key.create_sort_exp -> string
 *
 * Wrapper of SlZDataManagerSortKey::createSortExp(char*, ulong).
 * Return bytearray representation of sort exp.
 */
static VALUE sk_create_sort_exp(VALUE self)
{
    long buflen = 10 + 6 * rb_funcall(rb_iv_get(self, "@item_names"),
                                      rb_intern("size"),
                                      0);
    VALUE buf = rb_str_new("buf", buflen);

    SlZDataManagerSortKey* p_zsk = sk_get_zsk(self);
    ulong result_len = p_zsk->createSortExp(RSTRING(buf)->ptr,
                                            (ulong)buflen);
    VALUE result = rb_funcall(buf, rb_intern("slice"), 2, INT2FIX(0),
                              ULONG2NUM(result_len));
    return result;
}

/* DataBase */

namespace DataBase {
    enum DataBaseType {
        ADDRESSBOOK = 0,
        TODOLIST,
        DATEBOOK,
        MAIL_INBOX,
        MAIL_OUTBOX,
        MEMO,
    };
}

static const char* DATABASE_TYPES[] = {
    "ADDRESSBOOK",
    "TODOLIST",
    "DATEBOOK",
    "MAIL_INBOX",
    "MAIL_OUTBOX",
    "MEMO",
};

/*
 - It doesn't work; parameters are different between them.

static const QString (*db_file_name_funcs[])(bool) = {
    (const QString (*)(bool))SlZDataBase::addressbookFileName,
    (const QString (*)(bool))SlZDataBase::todolistFileName,
    (const QString (*)(bool))SlZDataBase::datebookFileName,
    (const QString (*)(bool))SlZDataBase::mailInboxFileName,
    (const QString (*)(bool))SlZDataBase::mailOutboxFileName,
    (const QString (*)(bool))SlZDataBase::memoFileName,
};

 - It works, but we cannot get memoItems via function.
   Because we cannot find the declaration of struct SlZdbItemInfo.

static SlZDataBaseItemInfo (*db_item_info_funcs[])() = {
    SlZDataBase::addressbookItems,
    SlZDataBase::todolistItems,
    SlZDataBase::datebookItems,
    SlZDataBase::maillistItems,
    SlZDataBase::maillistItems,
};

*/

static void zdb_free(void* p)
{
    SlZDataBase* p_zdb = static_cast<SlZDataBase*>(p);

    if (p_zdb->isOpened())
        p_zdb->close();

    delete p_zdb;
}

SlZDataBase* zdb_open(VALUE v_db_type, VALUE v_is_readonly, VALUE v_filename)
{
    int   db_type      = NUM2INT(v_db_type);
    bool  is_readonly  = RTEST(v_is_readonly);

    QString filename(RTEST(v_filename) ? StringValuePtr(v_filename) : "");

    SlZDataBase *p_zdb;

/*
 - Abandoned for the reasons described above.

    if (db_type < (sizeof(DATABASE_TYPES)/sizeof(char*))) {
	p_zdb = new SlZDataBase((*(db_file_name_funcs[db_type]))(false),
				(*(db_item_info_funcs[db_type]))(),
				0,
				is_readonly);
    }
    else {
        rb_raise(rb_eRuntimeError, "No such database type: %d", db_type);
    }
*/

    switch (db_type) {
    case DataBase::ADDRESSBOOK:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::addressbookFileName(),
                                SlZDataBase::addressbookItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::TODOLIST:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::todolistFileName(),
                                SlZDataBase::todolistItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::DATEBOOK:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::datebookFileName(),
                                SlZDataBase::datebookItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::MAIL_INBOX:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::mailInboxFileName(),
                                SlZDataBase::maillistItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::MAIL_OUTBOX:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::mailOutboxFileName(),
                                SlZDataBase::maillistItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::MEMO:

/* thanks to ZdbManager.cpp in zdbat-1.0.1 written by yakty */
    {
        VALUE ite = rb_funcall(cItemInfo, rb_intern("new"), 0);
        rb_funcall(ite, rb_intern("add_item_info"), 4,
                   INT2FIX(1),
                   rb_str_new2("CLAS"),
                   rb_const_get(cItemInfo, rb_intern("TEXT")),
                   rb_str_new2("title"));
        rb_funcall(ite, rb_intern("add_item_info"), 4,
                   INT2FIX(2),
                   rb_str_new2("MEM1"),
                   rb_const_get(cItemInfo, rb_intern("TEXT")),
                   rb_str_new2("contents"));
        rb_funcall(ite, rb_intern("add_item_info"), 4,
                   INT2FIX(3),
                   rb_str_new2("DATE"),
                   rb_const_get(cItemInfo, rb_intern("TIME")),
                   rb_str_new2("modified time"));

        SlZDataBaseItemInfo* p_sli = ite_get_sli(ite);

        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::memoFileName(),
                                *p_sli,
                                0,
                                is_readonly);
    }
/* end */
        break;
    default:
        rb_raise(rb_eRuntimeError, "No such database type: %d", db_type);
    }

    return p_zdb;
}

SlZDataBase* db_get_zdb(VALUE self)
{
    SlZDataBase* p_zdb;
    VALUE zdb = rb_iv_get(self, "zdb");

    if (NIL_P(zdb)) {
        rb_raise(rb_eRuntimeError, "Database not initialized.");
        return 0;
    }

    Data_Get_Struct(zdb, SlZDataBase, p_zdb);

    return p_zdb;
}

/*
 * Document-method: close
 * call-seq:
 *   database.close -> true or false
 *
 * Close the database.
 */
static VALUE db_close(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->close() ? Qtrue : Qfalse;
}

/*
 * Document-method: open
 * call-seq:
 *   database.open(is_readonly) -> true or false
 *
 * Open the database (again).
 * Notice: when DataBase instance is created, database is already opened.
 */
static VALUE db_open(VALUE self, VALUE is_readonly)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->open(RTEST(is_readonly)) ? Qtrue : Qfalse;
}

/*
 * Document-method: is_opened, opened?
 * call-seq:
 *   database.is_opened() -> true or false
 *   database.opened?() -> true or false
 *
 * Ask if the database is opened.
 */
static VALUE db_is_opened(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->isOpened() ? Qtrue : Qfalse;
}

/*
 * Document-method: is_closed, closed?
 * call-seq:
 *   database.is_closed() -> true or false
 *   database.closed?() -> true or false
 *
 * Opposite of DataBase#is_opened.
 */
static VALUE db_is_closed(VALUE self)
{
    return RTEST(rb_funcall(self, rb_intern("is_opened"), 0)) ? Qfalse : Qtrue;
}

VALUE db_init_item_info(VALUE self)
{
    VALUE item_info = rb_funcall(cItemInfo, rb_intern("new"), 0);

    SlZDataBase* p_zdb = db_get_zdb(self);

    size_t n_items(p_zdb->items().count());

    for (size_t i = 1; i < n_items; ++i) { // i = 0 is always CATEGORY
        QString name(p_zdb->itemName(i).left(4));
        if (!p_zdb->nameToFieldId(i))
            name = name.sprintf("NONAME%d", i);
        uint value_type(QChar(p_zdb->itemName(i).at(4)).latin1());
        rb_funcall(item_info, rb_intern("add_item_info"), 4,
                   INT2FIX(i), rb_str_new2((const char*)name),
                   INT2FIX(value_type), Qnil);
    }
    return item_info;
}

/*
 * Document-method: initialize
 * call-seq:
 *   DataBase.new(database_type, is_readonly = false, class_of_card = SimpleZDB::Card, filename = nil) -> database
 *
 * Open the database and return the DataBase object.
 * Possible values of database_type are:
 *
 *   DataBase::ADDRESSBOOK
 *   DataBase::TODOLIST
 *   DataBase::DATEBOOK
 *   DataBase::MAIL_INBOX
 *   DataBase::MAIL_OUTBOX
 *   DataBase::MEMO
 *
 * The third argument, class_of_card is used by DataBase#each method
 * in order to create an object as the block argument.
 * Normally it is SimpleZDB::Card or its subclass.
 * This class is expected that it has a constructor called
 * with two arguments, database and card_id.
 */

static VALUE db_initialize(VALUE self, VALUE args)
{
    VALUE db_type = rb_ary_shift(args);

    if (NIL_P(db_type)) {
        rb_warn("Database type not specified.");
        return self;
    }
    
    VALUE is_readonly = rb_ary_shift(args);
    VALUE class_of_card  = rb_ary_shift(args);
    VALUE filename    = rb_ary_shift(args);
    
    if (NIL_P(class_of_card)) {
        class_of_card = cCard;
    }
    rb_iv_set(self, "@class_of_card", class_of_card);

    SlZDataBase* p_zdb = zdb_open(db_type, is_readonly, filename);

    if (!p_zdb) {
        rb_raise(rb_eRuntimeError, "Open database failed.");
        return self;
    }

    VALUE zdb = Data_Wrap_Struct(self, 0, zdb_free, p_zdb);

    rb_iv_set(self, "zdb", zdb);

    rb_iv_set(self, "@item_info", db_init_item_info(self));

    return self;
}

/*
 * Document-method: open
 * call-seq:
 *   DataBase.open(database_type, is_readonly = false, class_of_card = SimpleZDB::Card, filename = nil) -> database
 *   DataBase.open(database_type, is_readonly = false, class_of_card = SimpleZDB::Card, filename = nil) {|database| ... }
 *
 * Open the database specified by the parameters.
 * See also DataBase.new.
 * If block given, open database, provide it as the block parameter,
 * execute the block, and close the database.
 */
static VALUE db_s_open(int argc, VALUE* argv, VALUE klass)
{
    VALUE obj = rb_funcall2(klass, rb_intern("new"), argc, argv);
    if (NIL_P(obj))
	return Qnil;
    if (rb_block_given_p()) {
	return rb_ensure(RUBY_METHOD_FUNC(rb_yield),
			 obj,
			 RUBY_METHOD_FUNC(db_close),
			 obj);
    }
    return obj;
}

/*
 * Document-method: size
 * call-seq:
 *   database.size -> fixnum
 */

static VALUE db_size(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return INT2FIX(p_zdb->count());
}

/*
 * Document-method: each
 * call-seq:
 *     database.each {|card| ... } -> database
 *
 * the class of the block argument is determined by the
 * instance variable @class_of_card, which can be accessed by
 * database.class_of_card method. Normally it is SimpleZDB::Card.
 */

static VALUE db_each(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    VALUE card;

    for (CardId card_id = p_zdb->firstCard();
         card_id != 0;
         card_id = p_zdb->nextCard(card_id)) {
        card = rb_funcall(rb_iv_get(self, "@class_of_card"),
                          rb_intern("new"),
                          2,
                          self,
                          ULONG2NUM(card_id));
        rb_yield(card);
    }
    return self;
}

/*
 * Document-method: read_time_field
 * call-seq:
 *   database.read_time_field(field, card_id) -> datetime or date
 *
 * (Normally not used directly. Called from DataBase#read_item.)
 * Return the date/time value specified by the field (i.e.,
 * return value of ItemInfo#field(name)) of the card card_id.
 */
static VALUE db_read_time_field(VALUE self, VALUE v_field, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    int    field   = NUM2INT(v_field);
    CardId card_id = NUM2ULONG(v_card_id);

    QDateTime q_datetime = p_zdb->readTimeField(field, card_id); // UTC
    QDate     q_date     = q_datetime.date();
    QTime     q_time     = q_datetime.time();

    VALUE val = Qnil;

    if (q_datetime.isValid()) {
        val = rb_funcall(cDateTime, rb_intern("new"), 7,
                         INT2FIX(q_date.year()),
                         INT2FIX(q_date.month()),
                         INT2FIX(q_date.day()),
                         INT2FIX(q_time.hour()),
                         INT2FIX(q_time.minute()),
                         INT2FIX(q_time.second()),
                         INT2FIX(0));
    }
    else if (q_date.isValid()) {
        val = rb_funcall(cDate, rb_intern("new"), 3,
                         INT2FIX(q_date.year()),
                         INT2FIX(q_date.month()),
                         INT2FIX(q_date.day()));
    }

    return val;
}

/*
 * Document-method: read_categories
 * call-seq:
 *   database.read_categories(card_id) -> array
 *
 * (Normally not used directly. Called from DataBase#read_item.)
 * Return the categories of the card card_id.
 */

static VALUE db_read_categories(VALUE self, VALUE card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    VALUE category_set = rb_iv_get(self, "@category_set");

    if (NIL_P(category_set)) {
        category_set = rb_funcall(cCategorySet, rb_intern("new"), 0);
        rb_iv_set(self, "@category_set", category_set);
    }

    if (RTEST(rb_funcall(card_id, rb_intern("respond_to?"), 1,
                         rb_str_new2("card_id")))) {
        card_id = rb_funcall(card_id, rb_intern("card_id"), 0);
    }

    VALUE val = rb_ary_new();

    QArray<int> ids = p_zdb->readCategories(NUM2ULONG(card_id));
    for (QArray<int>::ConstIterator it = ids.begin(); it != ids.end(); ++it) {
        rb_ary_push(val,
                    rb_funcall(category_set, rb_intern("[]"), 1,
                               INT2FIX((*it))));
    }
    return val;
}

/*
 * Document-method: read_item
 * call-seq:
 *   database.read_item(key, card_id) -> object
 *
 * (Normally not used directly. Called from Card#[])
 * Return the value specified by key of the card card_id.
 */

static VALUE db_read_item(VALUE self, VALUE key, VALUE v_card_id)
{
    VALUE item_info = rb_iv_get(self, "@item_info");
    VALUE result = ite_include_p(item_info, key);

    if (!RTEST(result))
        return Qnil;

    int field      = NUM2INT(ite_field(item_info, key));
    int value_type = NUM2INT(ite_value_type(item_info, key));

    if (RTEST(rb_funcall(v_card_id, rb_intern("respond_to?"),
                         1, rb_str_new2("card_id")))) {
        v_card_id = rb_funcall(v_card_id, rb_intern("card_id"), 0);
    }

    SlZDataBase* p_zdb = db_get_zdb(self);

    CardId card_id = NUM2ULONG(v_card_id);

    VALUE val;

    switch (value_type) {
    case SlZDataBaseItemInfo::TEXT:
        val = rb_str_new2(p_zdb->readField(field, card_id).utf8());
        break;
    case SlZDataBaseItemInfo::UCHAR:
        val = INT2FIX(p_zdb->readUcharField(field, card_id));
        break;
    case SlZDataBaseItemInfo::USHORT:
        val = INT2FIX(p_zdb->readUshortField(field, card_id));
        break;
    case SlZDataBaseItemInfo::ULONG:
        val = INT2FIX(p_zdb->readUlongField(field, card_id));
        break;
    case SlZDataBaseItemInfo::TIME:
        val = rb_funcall(self, rb_intern("read_time_field"), 2,
                         INT2FIX(field), ULONG2NUM(card_id));
        break;
    case SlZDataBaseItemInfo::CATEGORY:
        val = rb_funcall(self, rb_intern("read_categories"), 1,
                         ULONG2NUM(card_id));
        break;
    case SlZDataBaseItemInfo::BINARY:
    case SlZDataBaseItemInfo::COLOR:
    default:
        val = Qnil;
//        rb_raise(rb_eRuntimeError, "Item value_type %d not yet supported.", item_type);
    }
    return val;
}

/*
 * Document-method: write_time_field
 * call-seq:
 *   database.write_time_field(field, val) -> true or false
 *
 * (Normally not used directly. Called from DataBase#write_item)
 * Write the date/time value specified by the field (i.e.,
 * return value of the ItemInfo#field(key)) of the current card.
 */
static VALUE db_write_time_field(VALUE self, VALUE v_field, VALUE val)
{
    int field = NUM2INT(v_field);

    VALUE result = Qfalse;

    SlZDataBase* p_zdb = db_get_zdb(self);

    if (RTEST(rb_funcall(val, rb_intern("respond_to?"), 1,
                         rb_str_new2("hour")))) {
        val = rb_funcall(val, rb_intern("new_offset"), 0); // set UTC

        QDate q_date(NUM2INT(rb_funcall(val, rb_intern("year"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("month"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("day"), 0)));
        QTime q_time(NUM2INT(rb_funcall(val, rb_intern("hour"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("min"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("sec"), 0)));

        QDateTime q_datetime(q_date, q_time);
        result = p_zdb->writeField(field, q_datetime)
            ? Qtrue : Qfalse;
    }
    else {
        QDate q_date(NUM2INT(rb_funcall(val, rb_intern("year"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("month"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("day"), 0)));
        result = p_zdb->writeField(field, q_date)
            ? Qtrue : Qfalse;
    }
    return result;
}        

/*
 * Document-method: write_item
 * call-seq:
 *   database.write_item(key, val) -> true or false
 *
 * (Normally not used directly. Called from Card#save.)
 * Write the value specified by the key of the current card.
 */
static VALUE db_write_item(VALUE self, VALUE key, VALUE val)
{
    VALUE item_info = rb_iv_get(self, "@item_info");

    if (!RTEST(ite_include_p(item_info, key)))
        return Qnil;

    int field      = NUM2INT(ite_field(item_info, key));
    int value_type = NUM2INT(ite_value_type(item_info, key));

    VALUE result = Qfalse;

    SlZDataBase* p_zdb = db_get_zdb(self);

    switch (value_type) {
    case SlZDataBaseItemInfo::TEXT:
        {
            const char* c_val = StringValuePtr(val);
            result = p_zdb->writeField(field, QString::fromUtf8(c_val))
                ? Qtrue : Qfalse;
        }
        break;
    case SlZDataBaseItemInfo::UCHAR:
        result = p_zdb->writeField(field, (uchar)NUM2INT(val))
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::USHORT:
        result = p_zdb->writeField(field, (ushort)NUM2INT(val))
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::ULONG:
        result = p_zdb->writeField(field, (ulong)NUM2ULONG(val))
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::TIME:
        result = rb_funcall(self, rb_intern("write_time_field"),
                            2, INT2FIX(field), val)
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::CATEGORY: // must be done afterwards
        result = Qtrue;
    case SlZDataBaseItemInfo::BINARY:
    case SlZDataBaseItemInfo::COLOR:
    default:
        ;
//      p_zdb->cancelEditCard();
//      rb_raise(rb_eRuntimeError, "Item value_type %d not yet supported.", type);

    }
    return result;
}

/*
 * Document-method: update_categories
 * call-seq:
 *   database.upcate_categories(card_id, categories) -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Update the categories off the current card.
 * If any one of the categories does not exist in the CategorySet,
 * add it.
 */
static VALUE db_update_categories(VALUE self, VALUE v_card_id, VALUE categories)
{
    VALUE result = Qfalse;

    VALUE category_set = rb_iv_get(self, "@category_set");
    if (NIL_P(category_set)) {
        category_set = rb_funcall(cCategorySet, rb_intern("new"), 0);
        rb_iv_set(self, "@category_set", category_set);
    }

    VALUE ids = rb_funcall(category_set, rb_intern("get_ids_of_labels"),
                           1, categories);

    QArray<int> q_ids(NUM2INT(rb_funcall(ids, rb_intern("size"), 0)));
    for (QArray<int>::Iterator it = q_ids.begin(); it != q_ids.end(); ++it) {
        *it = NUM2INT(rb_ary_shift(ids));
    }

    SlZDataBase* p_zdb = db_get_zdb(self);

    CardId card_id = NUM2ULONG(v_card_id);
    
    result = p_zdb->updateCategories(card_id, q_ids) ? Qtrue : Qfalse;

    return result;
}

/*
 * Document-method: start_edit_card
 * call-seq:
 *   database.start_edit_card(card_id) -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Start editing the card. You should call this before writing
 * items.
 */
static VALUE db_start_edit_card(VALUE self, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->startEditCard(NUM2ULONG(v_card_id)) ? Qtrue : Qfalse;
}

/*
 * Document-method: finish_edit_card
 * call-seq:
 *   database.finish_edit_card(card_id) -> fixnum or false
 *
 * (Normally not used directly. Called from Card.)
 * Finish editing the card. You should call this after writing
 * items.
 * Return the new card_id.
 */
static VALUE db_finish_edit_card(VALUE self, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);
    CardId card_id(NUM2ULONG(v_card_id));

    bool result = p_zdb->finishEditCard(&card_id);

    if (result)
        return ULONG2NUM(card_id);
    else
        return Qfalse;
}

/*
 * Document-method: cancel_edit_card
 * call-seq:
 *   database.cancel_edit_card -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Cancel editing the card.
 */
static VALUE db_cancel_edit_card(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->cancelEditCard() ? Qtrue : Qfalse;
}

/*
 * Document-method: delete_card
 * call-seq:
 *   database.delete_card -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Delete the card.
 */
static VALUE db_delete_card(VALUE self, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    CardId card_id(NUM2ULONG(v_card_id));

    return p_zdb->deleteCard(&card_id) ? Qtrue : Qfalse;
}

/* Document-method: sort!
 * call-seq:
 *   database.sort!(key = nil, is_ascendent = true) -> database
 *
 * Sort cards by key. This method affects the result of each.
 * If key is invalid, do nothing.
 * If key is nil, sort by the default index.
 * Throw RuntimeError when sort failed.
 */
static VALUE db_sort_bang(VALUE self, VALUE args)
{
    VALUE key            = rb_ary_shift(args);
    VALUE v_is_ascendent = rb_ary_shift(args);

    int field = 0;

    VALUE item_info = rb_iv_get(self, "@item_info");

    if (!NIL_P(key)) {
        if (!RTEST(ite_include_p(item_info, key)))
            return self;

        field = NUM2INT(ite_field(item_info, key));
    }

    if (NIL_P(v_is_ascendent)) {
	v_is_ascendent = Qtrue;
    }

    bool is_ascendent = RTEST(v_is_ascendent);

    SlZDataBase* p_zdb = db_get_zdb(self);

    bool result = true;

    if (NIL_P(key)) {
        p_zdb->sort(is_ascendent);
    }
    else {
        result = p_zdb->sortField(field, is_ascendent);
    }

    if (!result) {
	rb_raise(rb_eRuntimeError, "Sort failed");
    }
    return self;
}

/*
 * Document-method: sort_exp!
 * call-seq:
 *   database.sort_exp!(exp, is_ascendent = true) -> database
 *
 * Wrapper of SlZDataManager::sort(const char*, int, bool).
 * The format of exp is undocumented.
 * If failed, raise runtime error.
 */
static VALUE db_sort_exp_bang(VALUE self, VALUE args)
{
    int argc = NUM2INT(rb_funcall(args, rb_intern("size"), 0));
    VALUE orig_exp       = rb_ary_shift(args);
    VALUE v_is_ascendent = rb_ary_shift(args);

    if (NIL_P(orig_exp)) {
        rb_raise(rb_eArgError,
                 "Arguments of sort_exp! too few (%d of 1 or 2)",
                 argc);
        return self;
    }

    VALUE exp = StringValue(orig_exp);

    if (NIL_P(v_is_ascendent)) {
        v_is_ascendent = Qtrue;
    }

    bool is_ascendent = RTEST(v_is_ascendent);

    SlZDataBase* p_zdb = db_get_zdb(self);

    bool result = p_zdb->sort(RSTRING(exp)->ptr,
                              RSTRING(exp)->len,
                              is_ascendent);
    if (!result) {
        rb_raise(rb_eRuntimeError,
                 "sort_exp!(\"%s\") failed.", RSTRING(exp)->ptr);
    }
    return self;
}

/*
 * Document-method: search_field
 * call-seq:
 *   database.search_field(str, field, search_mode, is_next = true) -> card or nil
 *
 * Wrapper of SlZDataBase::searchField(CardId*, QString, int, uchar, bool).
 * Valid search_mode values are undocumented.
 */
static VALUE db_search_field(VALUE self, VALUE args)
{
    int argc = NUM2INT(rb_funcall(args, rb_intern("size"), 0));
    if (argc < 3) {
        rb_raise(rb_eArgError, "Arguments of search_field too few (%d of 3 or 4)", argc);
        return Qnil;
    }

    VALUE v_str         = rb_ary_shift(args);
    VALUE v_field       = rb_ary_shift(args);
    VALUE v_search_mode = rb_ary_shift(args);
    VALUE v_is_next     = rb_ary_shift(args);
    if (NIL_P(v_is_next)) {
        v_is_next = Qtrue;
    }

    SlZDataBase* p_zdb = db_get_zdb(self);
    CardId card_id = 0;
    QString q_str(QString::fromUtf8(StringValuePtr(v_str)));
    bool result = p_zdb->searchField(&card_id,
                                     q_str,
                                     NUM2INT(v_field),
                                     (uchar)NUM2INT(v_search_mode),
                                     RTEST(v_is_next));
    return result
        ? rb_funcall(rb_iv_get(self, "@class_of_card"),
                     rb_intern("new"),
                     2,
                     self,
                     ULONG2NUM(card_id))
        : Qnil;
}

/*
 * Document-method: search_cond
 * call-seq:
 *   database.search_cond(cond, is_next, attr_mask, attr_cond, attr_flag, idx_name = nil) -> card or nil
 *   database.search_cond(cond, is_next, idx_name = nil) -> card or nil
 *
 * Wrapper of SlZDataManager::search(CardId*, const char*, bool, uchar, uchar, uchar, const char*)
 * and SlZDataManager::search(CardId*, const char*, bool, const char*).
 * The format of cond is undocumented.
 */
static VALUE db_search_cond(VALUE self, VALUE args)
{
    int argc = NUM2INT(rb_funcall(args, rb_intern("size"), 0));
    if (!((argc == 2) || (argc == 3) || (argc == 5) || (argc == 6))) {
        rb_raise(rb_eArgError,
                 "Number of argument not match (%d of 2, 3, 5, or 6)",
                 argc);
        return Qnil;
    }
    VALUE v_cond      = rb_ary_shift(args);
    VALUE v_is_next   = rb_ary_shift(args);
    VALUE v_attr_mask = INT2FIX(0);
    VALUE v_attr_cond = INT2FIX(0);
    VALUE v_attr_flag = INT2FIX(0);
    if (argc >= 5) {
        v_attr_mask = rb_ary_shift(args);
        v_attr_cond = rb_ary_shift(args);
        v_attr_flag = rb_ary_shift(args);
    }

    VALUE v_idx_name  = rb_ary_shift(args);

    SlZDataBase* p_zdb = db_get_zdb(self);

    CardId card_id = 0;

    bool result = false;
    if (argc >= 5) {
        result = p_zdb->search(&card_id,
                               StringValuePtr(v_cond),
                               RTEST(v_is_next),
                               (uchar)NUM2INT(v_attr_mask),
                               (uchar)NUM2INT(v_attr_cond),
                               (uchar)NUM2INT(v_attr_flag),
                               RTEST(v_idx_name)
                               ? StringValuePtr(v_idx_name)
                               : 0);
    }
    else {
        result = p_zdb->search(&card_id,
                               StringValuePtr(v_cond),
                               RTEST(v_is_next),
                               RTEST(v_idx_name)
                               ? StringValuePtr(v_idx_name)
                               : 0);
    }
    return result
        ? rb_funcall(rb_iv_get(self, "@class_of_card"),
                     rb_intern("new"),
                     2,
                     self,
                     ULONG2NUM(card_id))
        : Qnil;
}

static const char* FILTER_FOR[] = {
    "FOR_CURRENT_INDEX",
    "FOR_MASTER_INDEX",
    "FOR_SUSPENDED_INDEX",
};

/* Document-method: filter
 * call-seq:
 *   database.filter(search_exp, sort_exp, filter_mode, is_suspend_index = false) -> true or false
 *
 * Wrapper of SlZDataManager::filter(const char*, ulong, const char*, ulong, FilterFor, bool = false).
 * search_exp and sort_exp are undocumented.
 * Possible filter_mode values are:
 *
 *   DataBase::FOR_CURRENT_INDEX
 *   DataBase::FOR_MASTER_INDEX
 *   DataBase::FOR_SUSPENDED_INDEX
 */

static VALUE db_filter(VALUE self, VALUE args)
{
    int argc = NUM2INT(rb_funcall(args, rb_intern("size"), 0));
    if (!((argc == 3) || (argc == 4))) {
        rb_raise(rb_eArgError, "Number of arguments too few (%d of 3 or 4)",
                 argc);
        return Qnil;
    }
    VALUE orig_search_exp    = rb_ary_shift(args);
    VALUE orig_sort_exp      = rb_ary_shift(args);
    VALUE filter_mode        = rb_ary_shift(args);
    VALUE is_suspended_index = rb_ary_shift(args);
    if (NIL_P(is_suspended_index)) {
        is_suspended_index = Qfalse;
    }
    VALUE search_exp = StringValue(orig_search_exp);
    VALUE sort_exp
        = NIL_P(orig_sort_exp)
            ? Qnil
            : StringValue(orig_sort_exp);

    SlZDataBase* p_zdb = db_get_zdb(self);

    bool result = p_zdb->filter(RSTRING(search_exp)->ptr,
                                RSTRING(search_exp)->len,
                                RTEST(sort_exp) ? RSTRING(sort_exp)->ptr : 0,
                                RTEST(sort_exp) ? RSTRING(sort_exp)->len : 0,
                                (SlZDataManager::FilterFor)NUM2INT(filter_mode),
                                RTEST(is_suspended_index));

    return result ? Qtrue : Qfalse;
}

/*
 * Document-method: unfilter
 * call-seq:
 *   database.unfilter(is_resume_index = true) -> true or false
 *
 * Wrapper of SlZDataManager::unfilter(bool = true).
 * If failed, raise runtime error.
 */
static VALUE db_unfilter(VALUE self, VALUE args)
{
    int argc = NUM2INT(rb_funcall(args, rb_intern("size"), 0));
    if (argc > 1) {
        rb_raise(rb_eArgError, "Arguments too many (%d of 1).", argc);
        return Qfalse;
    }

    VALUE is_resume_index = rb_ary_shift(args);
    if (NIL_P(is_resume_index)) {
        is_resume_index = Qtrue;
    }

    SlZDataBase* p_zdb = db_get_zdb(self);
    return p_zdb->unfilter(RTEST(is_resume_index)) ? Qtrue : Qfalse;
}

/*
 * Document-method: kill_idx_resume
 * call-seq:
 *   database.kill_idx_resume -> database
 *
 * Wrapper of SlZDataManager::killIdxResume().
 */
static VALUE db_kill_idx_resume(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);
    p_zdb->killIdxResume();
    return self;
}

/*
 * Document-method: is_filtered, filtered?
 * call-seq:
 *   database.is_filtered -> true or false
 *   database.filtered? -> true or false
 *
 * Wrapper of SlZDataManager::isFiltered().
 */
static VALUE db_is_filtered(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);
    return p_zdb->isFiltered() ? Qtrue : Qfalse;
}

/*
 * Document-method: enable_filter
 * call-seq:
 *   database.enable_filter(category_id) -> true or false
 *
 * Wrapper of SlZDataManager::enableFilter(CategoryId).
 */
static VALUE db_enable_filter(VALUE self, VALUE category_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);
    return p_zdb->enableFilter(NUM2INT(category_id)) ? Qtrue : Qfalse;
}

/*
 * Document-method: disable_filter
 * call-seq:
 *   database.disable_filter(is_idx_delete) -> true or false
 *
 * Wrapper of SlZDataManager::disableFilter(bool)
 */
static VALUE db_disable_filter(VALUE self, VALUE is_idx_delete)
{
    SlZDataBase* p_zdb = db_get_zdb(self);
    return p_zdb->disableFilter(RTEST(is_idx_delete)) ? Qtrue : Qfalse;
}


/* for debug */
static VALUE db_inspect_fields(VALUE self)
{
    VALUE result = rb_ary_new();

    SlZDataBase* p_zdb = db_get_zdb(self);

    int count(p_zdb->items().count());

    VALUE used_field_ids = rb_ary_new();
    int field_id;

    for (int i = 0; i < count; ++i) {
        VALUE info = rb_ary_new();

                 field_id          = p_zdb->nameToFieldId(i);
        QString  item_name         ( p_zdb->itemName(i));
        DataType field_type        = p_zdb->readFieldType(i);
        QChar    another_item_type ( item_name.at(4));

        rb_ary_push(info, INT2FIX(field_id));
        rb_ary_push(info, rb_str_new2(item_name.left(4).latin1()));
        rb_ary_push(info, INT2FIX(field_type));
        rb_ary_push(info, INT2FIX(another_item_type.latin1()));

        rb_ary_push(used_field_ids, INT2FIX(field_id));

        rb_ary_push(result, info);
    }

    return result;
}

/* for debug */
static VALUE db_find_item_info(VALUE self, VALUE name)
{
    VALUE result = Qnil;

    SlZDataBase* p_zdb = db_get_zdb(self);

    QString q_name(QString::fromUtf8(StringValuePtr(name)));

    int item_id = p_zdb->nameToItemId(q_name);

    if (item_id > 0) {
        result = rb_ary_new();

        int      item_id   = p_zdb->nameToItemId(q_name);
        DataType item_type = p_zdb->readItemType(q_name);

        rb_ary_push(result, name);
        rb_ary_push(result, INT2FIX(item_id));
        rb_ary_push(result, INT2FIX(item_type));
    }

    return result;
}


/* CategorySet */

static void cs_free(void* p)
{
    delete static_cast<SlCategories*>(p);
}

static VALUE cs_alloc(VALUE klass)
{
    SlCategories* p_categories;
    VALUE obj;

    p_categories = new SlCategories();
    obj = Data_Wrap_Struct(klass, 0, cs_free, p_categories);
    return obj;
}

SlCategories* cs_get_slcats(VALUE self)
{
    SlCategories* p_slcats;
    Data_Get_Struct(self, SlCategories, p_slcats);
    return p_slcats;
}
/*
 * Document-method: []
 * call-seq:
 *   category_set[id] -> string
 *
 * Return the category label specified by id.
 */
static VALUE cs_get_label(VALUE self, VALUE id)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(p_slcats->label(NUM2INT(id)));

    return rb_str_new2(q_label.utf8());
}

/*
 * Document-method: id_of
 * call-seq:
 *   category_set.id_of(label) -> fixnum or nil.
 *
 * Return the id of the label.
 * Return nil if the label does not exist in the category set.
 */
static VALUE cs_get_id_of_label(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    int id = p_slcats->id(q_label);

    return id ? INT2FIX(id) : Qnil;
}

/*
 * Document-method: labels
 * call-seq:
 *   category_set.labels -> array
 */
static VALUE cs_labels(VALUE self)
{
    VALUE result = rb_ary_new();

    SlCategories* p_slcats = cs_get_slcats(self);

    QStringList q_labels(p_slcats->labels());
    for (QStringList::ConstIterator it = q_labels.begin();
	 it != q_labels.end();
	 ++it) {
	QString q_label(*it);
	VALUE label = rb_str_new2(q_label.utf8());
	rb_ary_push(result, label);
    }
    return result;
}

static VALUE cs_each_i(VALUE label, VALUE self)
{
    return rb_yield(label);
}

/*
 * Document-method: each
 * call-seq:
 *   category_set.each {|label| ... } -> category_set
 */
static VALUE cs_each(VALUE self)
{
    rb_iterate(RUBY_METHOD_FUNC(each),
               rb_funcall(self, rb_intern("labels"), 0),
	       RUBY_METHOD_FUNC(cs_each_i),
               self);
	       
    return self;
}
 
/*
 * Document-method: include?
 * call-seq:
 *   category_set.include?(label) -> true or false
 *
 * Return true if the label exist in the category set.
 */
static VALUE cs_include_p(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    return p_slcats->exists(q_label) ? Qtrue : Qfalse;
}

/*
 * Document-method: add
 * call-seq:
 *   category_set.add(label) -> fixnum
 *
 * Add the label to the category set. Return the id of the category.
 */
static VALUE cs_add(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    int new_id = p_slcats->addCategory(q_label);

    return INT2FIX(new_id);
}

/*
 * Document-method: delete
 * call-seq:
 *   category_set.delete(label) -> true or false
 *
 */
static VALUE cs_delete(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    bool result = p_slcats->removeCategory(q_label);

    return result ? Qtrue : Qfalse;
}

VALUE cs_get_ids_of_labels_i(VALUE label, VALUE args)
{
    VALUE result = rb_funcall(args, rb_intern("[]"), 1, INT2FIX(0));
    VALUE category_set = rb_funcall(args, rb_intern("[]"), 1, INT2FIX(1));

    if (!RTEST(rb_funcall(category_set, rb_intern("include?"), 1,
                          label))) {
        rb_funcall(category_set, rb_intern("add"), 1, label);
    }

    VALUE id = rb_funcall(category_set, rb_intern("id_of"), 1, label);

    rb_ary_push(result, id);
    return Qnil;
}

/*
 * Document-method: get_ids_of_labels
 * call-seq:
 *   category_set.get_ids_of_labels(labels) -> array
 */
static VALUE cs_get_ids_of_labels(VALUE self, VALUE labels)
{
    VALUE result = rb_ary_new();
    VALUE args = rb_ary_new();
    rb_ary_push(args, result);
    rb_ary_push(args, self);

    rb_iterate(RUBY_METHOD_FUNC(each),
               labels,
               RUBY_METHOD_FUNC(cs_get_ids_of_labels_i),
               args);
    return result;
}


extern "C" void Init_simplezdb()
{
    rb_require("date");
    cDate     = rb_funcall(rb_cObject, rb_intern("const_get"), 1,
                           rb_str_new2("Date"));
    cDateTime = rb_funcall(rb_cObject, rb_intern("const_get"), 1,
                           rb_str_new2("DateTime"));


    mSimpleZDB = rb_define_module("SimpleZDB");

    cDataBase = rb_define_class_under(mSimpleZDB,
				      "DataBase",
				      rb_cObject);
    rb_include_module(cDataBase, rb_mEnumerable);

    rb_define_attr(cDataBase,   "item_info", 1, 0);
    rb_define_attr(cDataBase,   "class_of_card", 1, 1);

    rb_define_singleton_method(cDataBase, "open", RUBY_METHOD_FUNC(db_s_open), -1);

    rb_define_method(cDataBase, "initialize", RUBY_METHOD_FUNC(db_initialize), -2);
    rb_define_method(cDataBase, "close",      RUBY_METHOD_FUNC(db_close), 0);
    rb_define_method(cDataBase, "open",       RUBY_METHOD_FUNC(db_open), 1);
    rb_define_method(cDataBase, "is_opened",  RUBY_METHOD_FUNC(db_is_opened), 0);
    rb_define_method(cDataBase, "opened?",    RUBY_METHOD_FUNC(db_is_opened), 0);
    rb_define_method(cDataBase, "is_closed",  RUBY_METHOD_FUNC(db_is_closed), 0);
    rb_define_method(cDataBase, "closed?",    RUBY_METHOD_FUNC(db_is_closed), 0);

    rb_define_method(cDataBase, "size",       RUBY_METHOD_FUNC(db_size), 0);
    rb_define_method(cDataBase, "length",     RUBY_METHOD_FUNC(db_size), 0);
    rb_define_method(cDataBase, "each",       RUBY_METHOD_FUNC(db_each), 0);


    rb_define_method(cDataBase, "read_time_field", RUBY_METHOD_FUNC(db_read_time_field), 2);
    rb_define_method(cDataBase, "read_categories", RUBY_METHOD_FUNC(db_read_categories), 1);
    rb_define_method(cDataBase, "read_item", RUBY_METHOD_FUNC(db_read_item), 2);
    rb_define_method(cDataBase, "write_time_field", RUBY_METHOD_FUNC(db_write_time_field), 2);
    rb_define_method(cDataBase, "write_item", RUBY_METHOD_FUNC(db_write_item), 2);
    rb_define_method(cDataBase, "update_categories", RUBY_METHOD_FUNC(db_update_categories), 2);
    rb_define_method(cDataBase, "start_edit_card",  RUBY_METHOD_FUNC(db_start_edit_card), 1);
    rb_define_method(cDataBase, "finish_edit_card", RUBY_METHOD_FUNC(db_finish_edit_card), 1);
    rb_define_method(cDataBase, "cancel_edit_card", RUBY_METHOD_FUNC(db_cancel_edit_card), 0);
    rb_define_method(cDataBase, "delete_card",      RUBY_METHOD_FUNC(db_delete_card), 1);
    rb_define_method(cDataBase, "sort!",            RUBY_METHOD_FUNC(db_sort_bang), -2);
    rb_define_method(cDataBase, "sort_exp!",        RUBY_METHOD_FUNC(db_sort_exp_bang), -2);
    rb_define_method(cDataBase, "search_field",     RUBY_METHOD_FUNC(db_search_field), -2);
    rb_define_method(cDataBase, "search_cond",      RUBY_METHOD_FUNC(db_search_cond), -2);
    rb_define_method(cDataBase, "filter",           RUBY_METHOD_FUNC(db_filter), -2);
    rb_define_method(cDataBase, "unfilter",         RUBY_METHOD_FUNC(db_unfilter), -2);
    rb_define_method(cDataBase, "kill_idx_resume",  RUBY_METHOD_FUNC(db_kill_idx_resume), 0);
    rb_define_method(cDataBase, "is_filtered",      RUBY_METHOD_FUNC(db_is_filtered), 0);
    rb_define_method(cDataBase, "filtered?",        RUBY_METHOD_FUNC(db_is_filtered), 0);
    rb_define_method(cDataBase, "enable_filter",    RUBY_METHOD_FUNC(db_enable_filter), 1);
    rb_define_method(cDataBase, "disable_filter",   RUBY_METHOD_FUNC(db_disable_filter), 1);

    rb_define_method(cDataBase, "inspect_fields", RUBY_METHOD_FUNC(db_inspect_fields), 0);
    rb_define_method(cDataBase, "find_item_info", RUBY_METHOD_FUNC(db_find_item_info), 1);

    size_t db_types_size(sizeof(DATABASE_TYPES)/sizeof(char*));
    for (size_t i = 0; i < db_types_size; ++i)
        rb_iv_set(cDataBase, DATABASE_TYPES[i], INT2FIX(i));

    size_t db_filter_for_size(sizeof(FILTER_FOR)/sizeof(char*));
    for (size_t i = 0; i < db_filter_for_size; ++i)
        rb_iv_set(cDataBase, FILTER_FOR[i], INT2FIX(i));

    rb_require("simplezdb/card");
    cCard     = rb_funcall(mSimpleZDB, rb_intern("const_get"), 1,
                           rb_str_new2("Card"));

/*
    rb_require("simplezdb/itemattr");
    cItemAttr = rb_funcall(mSimpleZDB, rb_intern("const_get"), 1,
                           rb_str_new2("ItemAttr"));
*/
    cCategorySet = rb_define_class_under(mSimpleZDB,
					 "CategorySet",
					 rb_cObject);

    rb_include_module(cCategorySet, rb_mEnumerable);

    rb_define_alloc_func(cCategorySet, RUBY_METHOD_FUNC(cs_alloc));
    rb_define_method(cCategorySet, "[]",       RUBY_METHOD_FUNC(cs_get_label), 1);
    rb_define_method(cCategorySet, "id_of",    RUBY_METHOD_FUNC(cs_get_id_of_label), 1);
    rb_define_method(cCategorySet, "labels",   RUBY_METHOD_FUNC(cs_labels), 0);
    rb_define_method(cCategorySet, "each",     RUBY_METHOD_FUNC(cs_each), 0);
    rb_define_alias(cCategorySet, "to_ary", "to_a");
    rb_define_method(cCategorySet, "include?", RUBY_METHOD_FUNC(cs_include_p), 1);
    rb_define_method(cCategorySet, "add",      RUBY_METHOD_FUNC(cs_add), 1);
    rb_define_method(cCategorySet, "delete",   RUBY_METHOD_FUNC(cs_delete), 1);
    rb_define_method(cCategorySet, "get_ids_of_labels", RUBY_METHOD_FUNC(cs_get_ids_of_labels), 1);

    cItemInfo = rb_define_class_under(mSimpleZDB,
                                      "ItemInfo",
                                      rb_cObject);
    rb_include_module(cItemInfo, rb_mEnumerable);
    rb_define_alloc_func(cItemInfo, RUBY_METHOD_FUNC(sli_alloc));
    rb_define_method(cItemInfo, "initialize", RUBY_METHOD_FUNC(ite_initialize), 0);
    rb_define_method(cItemInfo, "add_item_info", RUBY_METHOD_FUNC(ite_add_item_info), -2);
    rb_define_method(cItemInfo, "field",         RUBY_METHOD_FUNC(ite_field), 1);
    rb_define_method(cItemInfo, "value_type",    RUBY_METHOD_FUNC(ite_value_type), 1);
    rb_define_method(cItemInfo, "name",          RUBY_METHOD_FUNC(ite_name), 1);
    rb_define_method(cItemInfo, "size",          RUBY_METHOD_FUNC(ite_size), 0);
    rb_define_method(cItemInfo, "include?",      RUBY_METHOD_FUNC(ite_include_p), 1);
    rb_define_method(cItemInfo, "each",          RUBY_METHOD_FUNC(ite_each), 0);
    rb_define_attr(cItemInfo, "names", 1, 0);

    size_t ite_types_size(sizeof(ITEMINFO_TYPES)/sizeof(char*));
    for (size_t i = 0; i < ite_types_size; ++i)
        rb_iv_set(cItemInfo, ITEMINFO_TYPES[i], INT2FIX(i));

    cSortKey = rb_define_class_under(mSimpleZDB, "SortKey", rb_cObject);
    rb_define_alloc_func(cSortKey, RUBY_METHOD_FUNC(zsk_alloc));
    rb_define_method(cSortKey, "initialize", RUBY_METHOD_FUNC(sk_initialize), 0);
    rb_define_method(cSortKey, "add_sort_key", RUBY_METHOD_FUNC(sk_add_sort_key), -2);
    rb_define_method(cSortKey, "create_sort_exp", RUBY_METHOD_FUNC(sk_create_sort_exp), 0);
    rb_define_attr(cSortKey, "item_names", 1, 0);
    rb_define_attr(cSortKey, "is_string", 1, 0);
    rb_define_attr(cSortKey, "is_ascendent", 1, 0);
}
