require 'test/unit'

$:.unshift('./lib')

require 'simplezdb'
require 'date'

include SimpleZDB

module SimpleZDB
  class Contact < Card
  end
  class Task < Card
  end
  class Event < Card
  end
end

class TC_DataBase < Test::Unit::TestCase

  def test_s_new
    db = DataBase.new
    assert_instance_of(DataBase, db)
    assert_nil(db.item_info)
  end

  def test_s_new2
    db = DataBase.new
    assert_raises(RuntimeError) { db.close }
  end

  def test_s_new3
    db = DataBase.new(DataBase::ADDRESSBOOK)
    assert_nothing_raised { db.close }
  end

  def test_s_open
    DataBase.open(DataBase::ADDRESSBOOK) {|db|
      assert_instance_of(DataBase, db)
    }
  end

  def test_size
    db = DataBase.new(DataBase::TODOLIST)
    assert_kind_of(Integer, db.size)
    assert(db.size > 0)
    db.close
  end

  def test_each
    db = DataBase.new(DataBase::MEMO)
    assert_nothing_raised {
      db.each {|card|
        assert_instance_of(Card, card)
        assert_equal(db, card.database)
        assert_kind_of(Integer, card.card_id)
        break
      }
    }
    db.close
  end

  def test_each_event
    db = DataBase.new(DataBase::DATEBOOK, true, Event)
    assert_nothing_raised {
      db.each {|event|
        assert_kind_of(Event, event)
        break
      }
    }
    db.close
  end

  def test_addressbook
    db = DataBase.new(DataBase::ADDRESSBOOK)
    assert_instance_of(DataBase, db)
    assert_nothing_raised { db.close }
  end

  def test_todolist
    db = DataBase.new(DataBase::TODOLIST)
    assert_instance_of(DataBase, db)
    assert_nothing_raised { db.close }
  end

  def test_datebook
    db = DataBase.new(DataBase::DATEBOOK)
    assert_instance_of(DataBase, db)
    assert_nothing_raised { db.close }
  end

  def test_mail_inbox
    db = DataBase.new(DataBase::MAIL_INBOX)
    assert_instance_of(DataBase, db)
    assert_nothing_raised { db.close }
  end

  def test_mail_outbox
    db = DataBase.new(DataBase::MAIL_OUTBOX)
    assert_instance_of(DataBase, db)
    assert_nothing_raised { db.close }
  end

  def test_memo
    db = DataBase.new(DataBase::MEMO)
    assert_instance_of(DataBase, db)
    assert_nothing_raised { db.close }
  end

  def test_item_info
    db = DataBase.new(DataBase::ADDRESSBOOK)
    item_info = db.item_info
    assert_instance_of(ItemInfo, item_info)
    assert(item_info.size > 0)

    item_info.each do |field, name, value_type|
      assert_kind_of(Integer, field)
      assert_instance_of(String, name)
      assert_kind_of(Integer, value_type)
    end
    db.close
  end

  def test_size
    db = DataBase.new(DataBase::MEMO)
    assert_kind_of(Integer, db.size)
    assert(db.size >= 0)
    db.close
  end

  def test_read_item
    db = DataBase.new(DataBase::ADDRESSBOOK)
    item_info = db.item_info
    db.each do |card|
      name = item_info.name(rand(item_info.size))
      item = db.read_item(name, card.card_id)
      assert_not_nil(item)
      break
    end
    db.close
  end

  def test_read_time_field
    db = DataBase.new(DataBase::DATEBOOK)
    item_info = db.item_info
    time_field = nil
    item_info.each do |field, name, value_type|
      if value_type == ItemInfo::TIME then
        time_field = field
        break
      end
    end
    assert_not_nil(time_field)
    db.each do |card|
      item = db.read_time_field(time_field, card.card_id)
      assert_kind_of(Date, item)
      break
    end
    db.close
  end

  def test_read_categories
    db = DataBase.new(DataBase::ADDRESSBOOK)
    db.each do |card|
      categories = db.read_categories(card.card_id)
      assert_instance_of(Array, categories)
      if categories.size > 0 then
        assert_instance_of(String, categories.first)
        break
      end
    end
    db.close
  end

  def test_sort_bang
    db = DataBase.new(DataBase::DATEBOOK)
    assert_nothing_raised { db.sort!("TIM1", false) }
    first = nil
    db.each do |event|
      unless first
        first = event
        next
      end
      assert(event["TIM1"] <= first["TIM1"])
      break
    end
    assert_nothing_raised { db.sort! }
  end

  def test_sort_exp_bang
    db = DataBase.new(DataBase::ADDRESSBOOK)
    result = false
    assert_nothing_raised { db.sort_exp!("hoge") }
    db.sort!
  end

  def test_search_field
    db = DataBase.new(DataBase::ADDRESSBOOK, true, Contact)
    assert_raises(ArgumentError) { db.search_field }
    full_field = db.item_info.field("FULL")
    result = nil
    assert_nothing_raised {
      result = db.search_field("a", full_field, 0x91, false)
    }
    if result then
      puts "search_field match: #{result}"
      assert_kind_of(Contact, result)
    else
      puts "search_field not match."
    end
    db.close
  end

  def test_search_cond
    db = DataBase.new(DataBase::ADDRESSBOOK, true, Contact)
    assert_raises(ArgumentError) { db.search_cond }
    result = nil
    assert_nothing_raised {
      result = db.search_cond("hoge", false)
    }
    if result then
      assert_kind_of(Contact, result)
      puts "search_cond match: #{result}"
    else
      puts "search_cond not match. But don't worry. Because we don't know the spec of search_cond."
    end
    db.close
  end

  def test_filter
    db = DataBase.new(DataBase::ADDRESSBOOK)
    assert_raises(ArgumentError) { db.filter }
    result = nil
    assert_nothing_raised {
      result = db.filter("a", nil, DataBase::FOR_CURRENT_INDEX, true)
    }
    if result then
      puts "Filter success!"
    else
      puts "Filter failed! But don't worry. Because we don't know the spec of filter."
    end
    assert_nothing_raised {
      db.kill_idx_resume
    }
    assert_nothing_raised {
      db.unfilter
    }
    assert(!db.is_filtered)
    db.close
  end

  def test_enable_disable_filter
    db = DataBase.new(DataBase::ADDRESSBOOK)
    category_set = CategorySet.new
    friend_id = category_set.id_of("Friend")
    result = nil
    assert_nothing_raised {
      result = db.enable_filter(friend_id)
    }
    if result then
      puts "Filter category Friend succeeded"
    else
      puts "Filter category Friend failed"
    end
    db.each do |card|
      assert(card.categories.include?("Friend"))
    end
    assert_nothing_raised {
      db.disable_filter(true)
    }
  end
end

class TC_Card < Test::Unit::TestCase
  def setup
    @db = DataBase.new(DataBase::DATEBOOK)
    @card = nil
    @db.each do |card|
      @card = card
      break
    end
  end

  def teardown
    @db.close
  end

  def test_card_id
    assert_kind_of(Integer, @card.card_id)
  end

  def test_database
    assert_instance_of(DataBase, @card.database)
  end

  def test_keys
    assert_instance_of(Array, @card.keys)
  end

  def test_aref
    key = @card.keys[rand(@card.keys.size)]
    assert_not_nil(@card[key])
  end

  def test_aset
    str_key = nil
    item_info = @card.database.item_info
    item_info.each do |field, name, value_type|
      if value_type == ItemInfo::TEXT then
        str_key = name
        break
      end
    end
    assert_not_nil(str_key)
    orig = @card[str_key]
    @card[str_key] = 'Foo'
    assert_equal(@card[str_key], 'Foo')
    @card[str_key] = orig
    assert_equal(@card[str_key], orig)
  end

  def test_each
    size = @card.database.item_info.size
    count = 0
    @card.each do |val|
      count += 1
    end
    assert_equal(size, count)
  end

  def test_to_ary
    assert_instance_of(Array, @card.to_a)
  end

  def test_to_hash
    assert_instance_of(Hash, @card.to_hash)
  end

  def test_categories
    assert_instance_of(Array, @card.categories)
    if @card.categories.size > 0 then
      assert_instance_of(String, @card.categories.first)
    end
  end

  def test_categories_assign
    @card.categories = ["Home"]
    assert_equal(["Home"], @card.categories)
  end
end

class TC_CategorySet < Test::Unit::TestCase
  def setup
    @cs = CategorySet.new
  end

  def test_to_a
    categories = @cs.to_a
    assert_kind_of(Array, categories)
    if categories.size > 0 then
      assert_kind_of(String, categories.first)
    end
  end

  def test_include_p
    ["Friend", "God"].each do |label|
      if @cs.include?(label) then
        id = @cs.id_of(label)
        assert_equal(label, @cs[id])
      else
        assert_nil(@cs.id_of(label))
      end
    end
  end
end
